@@ -7,6 +7,12 @@
 
 static int crash_on_error = 0;
 
+/* Function for getting effective (specialized or not) frame handlers. */
+MVM_STATIC_INLINE MVMFrameHandler * MVM_frame_effective_handlers(MVMFrame *f) {
+    MVMSpeshCandidate *spesh_cand = f->spesh_cand;
+    return spesh_cand ? spesh_cand->handlers : f->static_info->body.handlers;
+}
+
 /* Maps ID of exception category to its name. */
 static const char * cat_name(MVMThreadContext *tc, MVMint32 cat) {
     switch (cat) {
@@ -83,22 +89,24 @@ static MVMint32 handler_can_handle(MVMFrame *f, MVMFrameHandler *fh, MVMint32 ca
         || ((category_mask & MVM_EX_CAT_CONTROL) && cat != MVM_EX_CAT_CATCH);
 }
 
-/* Looks through the handlers of a particular scope, and sees if one will
- * match what we're looking for. Returns 1 to it if so; if not,
- * returns 0. */
-static MVMint32 search_frame_handlers(MVMThreadContext *tc, MVMFrame *f,
-                                      MVMuint8 mode, MVMuint32 cat,
-                                      MVMObject *payload, LocatedHandler *lh) {
+/* Looks through the handlers of a particular frame, including inlines in
+ * dynamic scope, and sees if one will match what we're looking for. Returns
+ * 1 to it if so, and 0 if not; in the case 1 is returned the *lh will be
+ * populated with details of the located handler. Since upon inlining, the
+ * dynamic scope becomes lexical so far as the optimized bytecode is
+ * concerned, then this just needs a scan of the table without any further
+ * checks being needed. */
+static MVMint32 search_frame_handlers_dyn(MVMThreadContext *tc, MVMFrame *f,
+                                          MVMuint32 cat, MVMObject *payload,
+                                          LocatedHandler *lh) {
     MVMuint32  i;
     if (f->spesh_cand && f->spesh_cand->jitcode && f->jit_entry_label) {
         MVMJitHandler    *jhs = f->spesh_cand->jitcode->handlers;
-        MVMFrameHandler  *fhs = f->effective_handlers;
+        MVMFrameHandler  *fhs = MVM_frame_effective_handlers(f);
         MVMint32 num_handlers = f->spesh_cand->jitcode->num_handlers;
         void         **labels = f->spesh_cand->jitcode->labels;
         void       *cur_label = f->jit_entry_label;
         for (i = 0; i < num_handlers; i++) {
-            if (mode == MVM_EX_THROW_LEX && fhs[i].inlined_and_not_lexical)
-                continue;
             if (!handler_can_handle(f, &fhs[i], cat, payload))
                 continue;
             if (cur_label >= labels[jhs[i].start_label] &&
@@ -117,11 +125,9 @@ static MVMint32 search_frame_handlers(MVMThreadContext *tc, MVMFrame *f,
         if (f == tc->cur_frame)
             pc = (MVMuint32)(*tc->interp_cur_op - *tc->interp_bytecode_start);
         else
-            pc = (MVMuint32)(f->return_address - f->effective_bytecode);
+            pc = (MVMuint32)(f->return_address - MVM_frame_effective_bytecode(f));
         for (i = 0; i < num_handlers; i++) {
-            MVMFrameHandler  *fh = &f->effective_handlers[i];
-            if (mode == MVM_EX_THROW_LEX && fh->inlined_and_not_lexical)
-                continue;
+            MVMFrameHandler  *fh = &(MVM_frame_effective_handlers(f)[i]);
             if (!handler_can_handle(f, fh, cat, payload))
                 continue;
             if (pc >= fh->start_offset && pc <= fh->end_offset && !in_handler_stack(tc, fh, f)) {
@@ -133,10 +139,140 @@ static MVMint32 search_frame_handlers(MVMThreadContext *tc, MVMFrame *f,
     return 0;
 }
 
+/* Looks for lexically applicable handlers in the current frame, accounting
+ * for any inlines. The skip_first_inlinee flag indicates that we should skip
+ * looking until we have encountered an inline boundary indicator saying that
+ * we have crossed from an inlinee to its inliner's handlers; this is used to
+ * handle the THROW_LEX_CALLER mode. If we never encounter an inline boundary
+ * when skip_first_inlinee is true then we'll always return 0.
+ *
+ * If skip_first_inlinee is false or we already saw an inline boundary, then
+ * we start looking for a matching handler. If one is found before seeing
+ * another inline boundary, then it is applicable; the data pointed to by lh
+ * will be updated with the frame handler details and 1 will be returned.
+ *
+ * Upon reaching an (or, in the case of skip_first_inlinee, a second) inline
+ * boundary indicator, there are two cases that apply. We take the inline
+ * that we are leaving, and look up the code ref using the code_ref_reg in
+ * the inline. We know that we can never inline a frame that was closed over
+ * (due to capturelex or takeclosure being marked :noinline). Thus, either:
+ * 1. The outer of the inlinee is actually the current frame f. In this case,
+ *    we skip all inlined handlers and just consider those of f itself.
+ * 2. The next frame we should search in is the ->outer of the inlinee, and
+ *    thus all the rest of the handlers in this frame should be ignored. In
+ *    this case, the MVMFrame **next_outer will be populated with a pointer
+ *    to that frame.
+ *
+ * The skip_all_inlinees flag is set once we are below the frame on the stack
+ * to where the search started. Again, this is because a frame that did a
+ * lexical capture may not be inlined, so we only need to consider the topmost
+ * frame's handlers, not anything it might have inlined into it.
+ */
+static MVMint32 search_frame_handlers_lex(MVMThreadContext *tc, MVMFrame *f,
+                                          MVMuint32 cat, MVMObject *payload,
+                                          LocatedHandler *lh,
+                                          MVMuint32 *skip_first_inlinee,
+                                          MVMuint32 skip_all_inlinees,
+                                          MVMFrame **next_outer) {
+    MVMuint32 i;
+    MVMuint32 skipping = *skip_first_inlinee;
+    MVMFrameHandler *fhs = MVM_frame_effective_handlers(f);
+    if (f->spesh_cand && f->spesh_cand->jitcode && f->jit_entry_label) {
+        MVMJitHandler    *jhs = f->spesh_cand->jitcode->handlers;
+        MVMint32 num_handlers = f->spesh_cand->jitcode->num_handlers;
+        void         **labels = f->spesh_cand->jitcode->labels;
+        void       *cur_label = f->jit_entry_label;
+        for (i = 0; i < num_handlers; i++) {
+            MVMFrameHandler *fh = &(fhs[i]);
+            if (skip_all_inlinees && fh->inlinee >= 0)
+                continue;
+            if (fh->category_mask == MVM_EX_INLINE_BOUNDARY) {
+                if (cur_label >= labels[jhs[i].start_label] &&
+                        cur_label <= labels[jhs[i].end_label]) {
+                    if (skipping) {
+                        skipping = 0;
+                        *skip_first_inlinee = 0;
+                    }
+                    else {
+                        MVMuint16 cr_reg = f->spesh_cand->inlines[fh->inlinee].code_ref_reg;
+                        MVMFrame *inline_outer = ((MVMCode *)f->work[cr_reg].o)->body.outer;
+                        if (inline_outer == f) {
+                            skip_all_inlinees = 1;
+                        }
+                        else {
+                            *next_outer = inline_outer;
+                            return 0;
+                        }
+                    }
+                }
+                continue;
+            }
+            if (skipping || !handler_can_handle(f, fh, cat, payload))
+                continue;
+            if (cur_label >= labels[jhs[i].start_label] &&
+                    cur_label <= labels[jhs[i].end_label] &&
+                    !in_handler_stack(tc, fh, f)) {
+                if (skipping && f->static_info->body.is_thunk)
+                    return 0;
+                lh->handler     = fh;
+                lh->jit_handler = &jhs[i];
+                return 1;
+            }
+        }
+    }
+    else {
+        MVMint32 num_handlers = f->spesh_cand
+            ? f->spesh_cand->num_handlers
+            : f->static_info->body.num_handlers;
+        MVMint32 pc;
+        if (f == tc->cur_frame)
+            pc = (MVMuint32)(*tc->interp_cur_op - *tc->interp_bytecode_start);
+        else
+            pc = (MVMuint32)(f->return_address - MVM_frame_effective_bytecode(f));
+        for (i = 0; i < num_handlers; i++) {
+            MVMFrameHandler *fh = &(fhs[i]);
+            if (skip_all_inlinees && fh->inlinee >= 0)
+                continue;
+            if (fh->category_mask == MVM_EX_INLINE_BOUNDARY) {
+                if (pc >= fh->start_offset && pc <= fh->end_offset) {
+                    if (skipping) {
+                        skipping = 0;
+                        *skip_first_inlinee = 0;
+                    }
+                    else {
+                        MVMuint16 cr_reg = f->spesh_cand->inlines[fh->inlinee].code_ref_reg;
+                        MVMFrame *inline_outer = ((MVMCode *)f->work[cr_reg].o)->body.outer;
+                        if (inline_outer == f) {
+                            skip_all_inlinees = 1;
+                        }
+                        else {
+                            *next_outer = inline_outer;
+                            return 0;
+                        }
+                    }
+                }
+                continue;
+            }
+            if (skipping || !handler_can_handle(f, fh, cat, payload))
+                continue;
+            if (pc >= fh->start_offset &&
+                    pc <= fh->end_offset &&
+                    !in_handler_stack(tc, fh, f)) {
+                if (skipping && f->static_info->body.is_thunk)
+                    return 0;
+                lh->handler = fh;
+                return 1;
+            }
+        }
+    }
+    return 0;
+}
+
 /* Searches for a handler of the specified category, relative to the given
  * starting frame, searching according to the chosen mode. */
 static LocatedHandler search_for_handler_from(MVMThreadContext *tc, MVMFrame *f,
         MVMuint8 mode, MVMuint32 cat, MVMObject *payload) {
+    MVMuint32 skip_first_inlinee = 0;
     LocatedHandler lh;
     lh.frame = NULL;
     lh.handler = NULL;
@@ -144,25 +280,39 @@ static LocatedHandler search_for_handler_from(MVMThreadContext *tc, MVMFrame *f,
     lh.handler_out_of_dynamic_scope = 0;
     switch (mode) {
         case MVM_EX_THROW_LEX_CALLER:
-            f = f->caller;
-            while (f && f->static_info->body.is_thunk)
-                f = f->caller;
-            /* And now we've gone down a caller, it's just lexical... */
-        case MVM_EX_THROW_LEX:
+            skip_first_inlinee = 1;
+        case MVM_EX_THROW_LEX: {
+            MVMint32 skip_all_inlinees = 0;
             while (f != NULL) {
-                if (search_frame_handlers(tc, f, MVM_EX_THROW_LEX, cat, payload, &lh)) {
+                MVMFrame *outer_from_inlinee = NULL;
+                if (search_frame_handlers_lex(tc, f, cat, payload, &lh, &skip_first_inlinee,
+                            skip_all_inlinees, &outer_from_inlinee)) {
                     if (in_caller_chain(tc, f))
                         lh.frame = f;
                     else
                         lh.handler_out_of_dynamic_scope = 1;
                     return lh;
                 }
-                f = f->outer;
+                if (skip_first_inlinee) {
+                    /* If this is still set, it means that the topmost frame
+                     * had no inlines, so we didn't already reach a chain of
+                     * outers to traverse. In this case, skip over any thunks
+                     * and continue the search. */
+                    skip_first_inlinee = 0;
+                    f = f->caller;
+                    while (f && f->static_info->body.is_thunk)
+                        f = f->caller;
+                }
+                else {
+                    f = outer_from_inlinee ? outer_from_inlinee : f->outer;
+                    skip_all_inlinees = 1;
+                }
             }
             return lh;
+        }
         case MVM_EX_THROW_DYN:
             while (f != NULL) {
-                if (search_frame_handlers(tc, f, mode, cat, payload, &lh)) {
+                if (search_frame_handlers_dyn(tc, f, cat, payload, &lh)) {
                     lh.frame = f;
                     return lh;
                 }
@@ -220,13 +370,9 @@ static void run_handler(MVMThreadContext *tc, LocatedHandler lh, MVMObject *ex_o
 
         /* Ensure we have an exception object. */
         if (ex_obj == NULL) {
-            MVMROOT(tc, cur_frame, {
-            MVMROOT(tc, lh.frame, {
-            MVMROOT(tc, payload, {
+            MVMROOT3(tc, cur_frame, lh.frame, payload, {
                 ex_obj = MVM_repr_alloc_init(tc, tc->instance->boot_types.BOOTException);
             });
-            });
-            });
             ((MVMException *)ex_obj)->body.category = category;
             MVM_ASSIGN_REF(tc, &(ex_obj->header), ((MVMException *)ex_obj)->body.payload, payload);
         }
@@ -244,11 +390,10 @@ static void run_handler(MVMThreadContext *tc, LocatedHandler lh, MVMObject *ex_o
 
         /* Set up special return to unwinding after running the
          * handler. */
-        cur_frame->return_value        = (MVMRegister *)&tc->last_handler_result;
-        cur_frame->return_type         = MVM_RETURN_OBJ;
-        cur_frame->special_return      = unwind_after_handler;
-        cur_frame->special_unwind      = cleanup_active_handler;
-        cur_frame->special_return_data = ah;
+        cur_frame->return_value = (MVMRegister *)&tc->last_handler_result;
+        cur_frame->return_type = MVM_RETURN_OBJ;
+        MVM_frame_special_return(tc, cur_frame, unwind_after_handler, cleanup_active_handler,
+            ah, NULL);
 
         /* Invoke the handler frame and return to runloop. */
         STABLE(handler_code)->invoke(tc, handler_code, MVM_callsite_get_common(tc, MVM_CALLSITE_ID_NULL_ARGS),
@@ -313,14 +458,15 @@ static void cleanup_active_handler(MVMThreadContext *tc, void *sr_data) {
     MVM_free(ah);
 }
 
-char * MVM_exception_backtrace_line(MVMThreadContext *tc, MVMFrame *cur_frame, MVMuint16 not_top) {
+char * MVM_exception_backtrace_line(MVMThreadContext *tc, MVMFrame *cur_frame,
+                                    MVMuint16 not_top, MVMuint8 *throw_address) {
     MVMString *filename = cur_frame->static_info->body.cu->body.filename;
     MVMString *name = cur_frame->static_info->body.name;
     /* XXX TODO: make the caller pass in a char ** and a length pointer so
      * we can update it if necessary, and the caller can cache it. */
     char *o = MVM_malloc(1024);
-    MVMuint8 *cur_op = not_top ? cur_frame->return_address : cur_frame->throw_address;
-    MVMuint32 offset = cur_op - cur_frame->effective_bytecode;
+    MVMuint8 *cur_op = not_top ? cur_frame->return_address : throw_address;
+    MVMuint32 offset = cur_op - MVM_frame_effective_bytecode(cur_frame);
     MVMBytecodeAnnotation *annot = MVM_bytecode_resolve_annotation(tc, &cur_frame->static_info->body,
                                         offset > 0 ? offset - 1 : 0);
 
@@ -332,11 +478,11 @@ char * MVM_exception_backtrace_line(MVMThreadContext *tc, MVMFrame *cur_frame, M
         : NULL;
 
     char *filename_c = filename
-	? MVM_string_utf8_encode_C_string(tc, filename)
-	: "<ephemeral file>";
+        ? MVM_string_utf8_encode_C_string(tc, filename)
+        : "<ephemeral file>";
     char *name_c = name
-	? MVM_string_utf8_encode_C_string(tc, name)
-	: "<anonymous frame>";
+        ? MVM_string_utf8_encode_C_string(tc, name)
+        : "<anonymous frame>";
 
     snprintf(o, 1024, " %s %s:%u  (%s:%s)",
         not_top ? "from" : "  at",
@@ -346,9 +492,9 @@ char * MVM_exception_backtrace_line(MVMThreadContext *tc, MVMFrame *cur_frame, M
         name_c
     );
     if (filename)
-	MVM_free(filename_c);
+        MVM_free(filename_c);
     if (name)
-	MVM_free(name_c);
+        MVM_free(name_c);
 
     if (tmp1)
         MVM_free(tmp1);
@@ -364,11 +510,15 @@ MVMObject * MVM_exception_backtrace(MVMThreadContext *tc, MVMObject *ex_obj) {
     MVMObject *arr = NULL, *annotations = NULL, *row = NULL, *value = NULL;
     MVMuint32 count = 0;
     MVMString *k_file = NULL, *k_line = NULL, *k_sub = NULL, *k_anno = NULL;
+    MVMuint8 *throw_address;
 
-    if (IS_CONCRETE(ex_obj) && REPR(ex_obj)->ID == MVM_REPR_ID_MVMException)
+    if (IS_CONCRETE(ex_obj) && REPR(ex_obj)->ID == MVM_REPR_ID_MVMException) {
         cur_frame = ((MVMException *)ex_obj)->body.origin;
-    else
+        throw_address = ((MVMException *)ex_obj)->body.throw_address;
+    }
+    else {
         MVM_exception_throw_adhoc(tc, "Op 'backtrace' needs an exception object");
+    }
 
     MVM_gc_root_temp_push(tc, (MVMCollectable **)&arr);
     MVM_gc_root_temp_push(tc, (MVMCollectable **)&annotations);
@@ -388,8 +538,8 @@ MVMObject * MVM_exception_backtrace(MVMThreadContext *tc, MVMObject *ex_obj) {
     arr = MVM_repr_alloc_init(tc, tc->instance->boot_types.BOOTArray);
 
     while (cur_frame != NULL) {
-        MVMuint8             *cur_op = count ? cur_frame->return_address : cur_frame->throw_address;
-        MVMuint32             offset = cur_op - cur_frame->effective_bytecode;
+        MVMuint8             *cur_op = count ? cur_frame->return_address : throw_address;
+        MVMuint32             offset = cur_op - MVM_frame_effective_bytecode(cur_frame);
         MVMBytecodeAnnotation *annot = MVM_bytecode_resolve_annotation(tc, &cur_frame->static_info->body,
                                             offset > 0 ? offset - 1 : 0);
         MVMint32              fshi   = annot ? (MVMint32)annot->filename_string_heap_index : -1;
@@ -447,11 +597,11 @@ MVMObject * MVM_exception_backtrace_strings(MVMThreadContext *tc, MVMObject *ex_
     arr = MVM_repr_alloc_init(tc, tc->instance->boot_types.BOOTArray);
     cur_frame = ex->body.origin;
 
-    MVMROOT(tc, arr, {
-    MVMROOT(tc, cur_frame, {
+    MVMROOT2(tc, arr, cur_frame, {
         MVMuint32 count = 0;
         while (cur_frame != NULL) {
-            char      *line     = MVM_exception_backtrace_line(tc, cur_frame, count++);
+            char *line = MVM_exception_backtrace_line(tc, cur_frame, count++,
+                ex->body.throw_address);
             MVMString *line_str = MVM_string_utf8_decode(tc, tc->instance->VMString, line, strlen(line));
             MVMObject *line_obj = MVM_repr_box_str(tc, tc->instance->boot_types.BOOTStr, line_str);
             MVM_repr_push_o(tc, arr, line_obj);
@@ -459,7 +609,6 @@ MVMObject * MVM_exception_backtrace_strings(MVMThreadContext *tc, MVMObject *ex_
             MVM_free(line);
         }
     });
-    });
 
     return arr;
 }
@@ -470,7 +619,8 @@ void MVM_dump_backtrace(MVMThreadContext *tc) {
     MVMuint32 count = 0;
     MVMROOT(tc, cur_frame, {
         while (cur_frame != NULL) {
-            char *line = MVM_exception_backtrace_line(tc, cur_frame, count++);
+            char *line = MVM_exception_backtrace_line(tc, cur_frame, count++,
+                *(tc->interp_cur_op));
             fprintf(stderr, "%s\n", line);
             MVM_free(line);
             cur_frame = cur_frame->caller;
@@ -604,7 +754,7 @@ void MVM_exception_throwobj(MVMThreadContext *tc, MVMuint8 mode, MVMObject *ex_o
 
     if (!ex->body.origin) {
         MVM_ASSIGN_REF(tc, &(ex->common.header), ex->body.origin, tc->cur_frame);
-        tc->cur_frame->throw_address = *(tc->interp_cur_op);
+        ex->body.throw_address = *(tc->interp_cur_op);
     }
 
     run_handler(tc, lh, ex_obj, 0, NULL);
@@ -640,7 +790,7 @@ void MVM_exception_resume(MVMThreadContext *tc, MVMObject *ex_obj) {
     target = ex->body.origin;
     if (!target)
         MVM_exception_throw_adhoc(tc, "This exception is not resumable");
-    if (target->special_return != unwind_after_handler)
+    if (!target->extra || target->extra->special_return != unwind_after_handler)
         MVM_exception_throw_adhoc(tc, "This exception is not resumable");
     if (!in_caller_chain(tc, target))
         MVM_exception_throw_adhoc(tc, "Too late to resume this exception");
@@ -652,8 +802,7 @@ void MVM_exception_resume(MVMThreadContext *tc, MVMObject *ex_obj) {
         MVM_exception_throw_adhoc(tc, "Can only resume the current exception");
 
     /* Clear special return handler; we'll do its work here. */
-    target->special_return = NULL;
-    target->special_unwind = NULL;
+    MVM_frame_clear_special_return(tc, target);
 
     /* Clear the current active handler. */
     ah = tc->active_handlers;
@@ -665,114 +814,10 @@ void MVM_exception_resume(MVMThreadContext *tc, MVMObject *ex_obj) {
     MVM_frame_unwind_to(tc, target, ex->body.resume_addr, 0, NULL);
 }
 
-static MVMObject* get_lexotic_for_handler_idx(MVMThreadContext *tc, MVMint32 handler_idx) {
-    MVMLexotic *lexotic;
-    MVMStaticFrame  *sf = tc->cur_frame->static_info;
-    /* See if we've got this lexotic cached; return it if so. */
-    if (sf->body.pool_index < tc->lexotic_cache_size) {
-        lexotic = tc->lexotic_cache[sf->body.pool_index];
-        if (lexotic && lexotic->body.handler_idx == handler_idx)
-            return (MVMObject *)lexotic;
-    }
-
-    /* Allocate lexotic object, set it up, and cache it. */
-    MVMROOT(tc, sf, {
-        lexotic = (MVMLexotic *)MVM_repr_alloc_init(tc, tc->instance->Lexotic);
-    });
-    lexotic->body.handler_idx = handler_idx;
-    MVM_ASSIGN_REF(tc, &(lexotic->common.header), lexotic->body.sf, sf);
-    if (sf->body.pool_index >= tc->lexotic_cache_size) {
-        MVMuint32 orig_size = tc->lexotic_cache_size;
-        tc->lexotic_cache_size = sf->body.pool_index + 1;
-        tc->lexotic_cache = orig_size
-            ? MVM_realloc(tc->lexotic_cache, tc->lexotic_cache_size * sizeof(MVMLexotic *))
-            : MVM_malloc(tc->lexotic_cache_size * sizeof(MVMLexotic *));
-        memset(tc->lexotic_cache + orig_size, 0,
-            (tc->lexotic_cache_size - orig_size) * sizeof(MVMLexotic *));
-    }
-    if (!tc->lexotic_cache[sf->body.pool_index])
-        tc->lexotic_cache[sf->body.pool_index] = lexotic;
-
-    return (MVMObject *)lexotic;
-}
-
-/* Creates a new lexotic. */
-MVMObject * MVM_exception_newlexotic(MVMThreadContext *tc, MVMuint32 offset) {
-    /* Locate handler associated with the specified label. */
-    MVMFrame       *f     = tc->cur_frame;
-    MVMStaticFrame *sf    = f->static_info;
-    MVMint32 handler_idx  = -1;
-    MVMint32 num_handlers = f->spesh_cand
-        ? f->spesh_cand->num_handlers
-        : sf->body.num_handlers;
-    MVMuint32 i;
-    for (i = 0; i < num_handlers; i++) {
-        if (f->effective_handlers[i].action == MVM_EX_ACTION_GOTO &&
-                f->effective_handlers[i].goto_offset == offset) {
-            handler_idx = i;
-            break;
-        }
-    }
-    if (handler_idx < 0)
-        MVM_exception_throw_adhoc(tc, "Label with no handler passed to newlexotic");
-    return get_lexotic_for_handler_idx(tc, handler_idx);
-}
-
-/* Creates a new lexotic from the JIT. The JIT doesn't have access to
- * the offset, so we can't find it from within there. */
-MVMObject * MVM_exception_newlexotic_from_jit(MVMThreadContext *tc, MVMint32 label) {
-    /* Locate handler associated with the specified label. */
-    MVMFrame       *f       = tc->cur_frame;
-    MVMint32 handler_idx    = -1;
-    MVMint32 num_handlers   = f->spesh_cand->jitcode->num_handlers;
-    MVMJitHandler *handlers = f->spesh_cand->jitcode->handlers;
-    MVMuint32 i;
-    for (i = 0; i < num_handlers; i++) {
-        if (f->effective_handlers[i].action == MVM_EX_ACTION_GOTO &&
-            handlers[i].goto_label == label) {
-            handler_idx = i;
-            break;
-        }
-    }
-    if (handler_idx < 0)
-        MVM_exception_throw_adhoc(tc, "Label with no handler passed to newlexotic");
-    return get_lexotic_for_handler_idx(tc, handler_idx);
-}
-
-
-/* Unwinds to a lexotic captured handler. */
-void MVM_exception_gotolexotic(MVMThreadContext *tc, MVMint32 handler_idx, MVMStaticFrame *sf) {
-    MVMFrame *f, *search;
-    f = NULL;
-    search = tc->cur_frame;
-    while (search) {
-        f = search;
-        while (f) {
-            if (f->static_info == sf)
-                break;
-            f = f->outer;
-        }
-        if (f)
-            break;
-        search = search->caller;
-    }
-    if (f && in_caller_chain(tc, f)) {
-        LocatedHandler lh;
-        lh.frame = f;
-        lh.handler = &(f->effective_handlers[handler_idx]);
-        if (f->spesh_cand && f->spesh_cand->jitcode)
-            lh.jit_handler = &(f->spesh_cand->jitcode->handlers[handler_idx]);
-        else
-            lh.jit_handler = NULL;
-        run_handler(tc, lh, NULL, MVM_EX_CAT_RETURN, NULL);
-    }
-    else {
-        MVM_exception_throw_adhoc(tc, "Too late to invoke lexotic return");
-    }
-}
-
 /* Panics and shuts down the VM. Don't do this unless it's something quite
- * unrecoverable.
+ * unrecoverable, and a thread context is either not available or stands a
+ * good chance of being too corrupt to print (or is not relevant information).
+ * Use MVM_oops in the case a thread context is available.
  * TODO: Some hook for embedders.
  */
 MVM_NO_RETURN
@@ -799,12 +844,15 @@ void MVM_panic_allocation_failed(size_t len) {
 MVM_NO_RETURN
 void MVM_oops(MVMThreadContext *tc, const char *messageFormat, ...) {
     va_list args;
+    fprintf(stderr, "MoarVM oops: ");
     va_start(args, messageFormat);
     vfprintf(stderr, messageFormat, args);
     va_end(args);
     fprintf(stderr, "\n");
     MVM_dump_backtrace(tc);
     fprintf(stderr, "\n");
+    if (tc->instance->jit_log_fh)
+        fflush(tc->instance->jit_log_fh);
     exit(1);
 }
 
@@ -839,7 +887,6 @@ MVM_NO_RETURN
 void MVM_exception_throw_adhoc_free_va(MVMThreadContext *tc, char **waste, const char *messageFormat, va_list args) {
     LocatedHandler lh;
     MVMException *ex;
-
     /* The current frame will be assigned as the thrower of the exception, so
      * force it onto the heap before we begin. */
     if (tc->cur_frame)
@@ -850,7 +897,8 @@ void MVM_exception_throw_adhoc_free_va(MVMThreadContext *tc, char **waste, const
     MVMROOT(tc, ex, {
         char      *c_message = MVM_malloc(1024);
         int        bytes     = vsnprintf(c_message, 1024, messageFormat, args);
-        MVMString *message   = MVM_string_utf8_decode(tc, tc->instance->VMString, c_message, bytes);
+        int        to_encode = bytes > 1024 ? 1024 : bytes;
+        MVMString *message   = MVM_string_utf8_decode(tc, tc->instance->VMString, c_message, to_encode);
         MVM_free(c_message);
 
         /* Clean up after ourselves to avoid leaking C strings. */
@@ -862,7 +910,7 @@ void MVM_exception_throw_adhoc_free_va(MVMThreadContext *tc, char **waste, const
         MVM_ASSIGN_REF(tc, &(ex->common.header), ex->body.message, message);
         if (tc->cur_frame) {
             ex->body.origin = tc->cur_frame;
-            tc->cur_frame->throw_address = *(tc->interp_cur_op);
+            ex->body.throw_address = *(tc->interp_cur_op);
         }
         else {
             ex->body.origin = NULL;
@@ -908,3 +956,44 @@ void MVM_exception_throw_adhoc_free_va(MVMThreadContext *tc, char **waste, const
 void MVM_crash_on_error(void) {
     crash_on_error = 1;
 }
+
+MVMint32 MVM_get_exception_category(MVMThreadContext *tc, MVMObject *ex) {
+    if (IS_CONCRETE(ex) && REPR(ex)->ID == MVM_REPR_ID_MVMException)
+        return ((MVMException *)ex)->body.category;
+    else
+        MVM_exception_throw_adhoc(tc, "getexcategory needs a VMException, got %s (%s)", REPR(ex)->name, MVM_6model_get_debug_name(tc, ex));
+}
+
+MVMObject * MVM_get_exception_payload(MVMThreadContext *tc, MVMObject *ex) {
+    MVMObject *result;
+    if (IS_CONCRETE(ex) && REPR(ex)->ID == MVM_REPR_ID_MVMException)
+        result = ((MVMException *)ex)->body.payload;
+    else
+        MVM_exception_throw_adhoc(tc, "getexpayload needs a VMException, got %s (%s)", REPR(ex)->name, MVM_6model_get_debug_name(tc, ex));
+    if (!result)
+        result = tc->instance->VMNull;
+    return result;
+}
+
+void MVM_bind_exception_payload(MVMThreadContext *tc, MVMObject *ex, MVMObject *payload) {
+    if (IS_CONCRETE(ex) && REPR(ex)->ID == MVM_REPR_ID_MVMException) {
+        MVM_ASSIGN_REF(tc, &(ex->header), ((MVMException *)ex)->body.payload,
+                payload);
+    }
+    else {
+        MVM_exception_throw_adhoc(tc, "bindexpayload needs a VMException, got %s (%s)", REPR(ex)->name, MVM_6model_get_debug_name(tc, ex));
+    }
+}
+
+void MVM_bind_exception_category(MVMThreadContext *tc, MVMObject *ex, MVMint32 category) {
+    if (IS_CONCRETE(ex) && REPR(ex)->ID == MVM_REPR_ID_MVMException)
+        ((MVMException *)ex)->body.category = category;
+    else
+        MVM_exception_throw_adhoc(tc, "bindexcategory needs a VMException, got %s (%s)", REPR(ex)->name, MVM_6model_get_debug_name(tc, ex));
+}
+void MVM_exception_returnafterunwind(MVMThreadContext *tc, MVMObject *ex) {
+    if (IS_CONCRETE(ex) && REPR(ex)->ID == MVM_REPR_ID_MVMException)
+        ((MVMException *)ex)->body.return_after_unwind = 1;
+    else
+        MVM_exception_throw_adhoc(tc, "exreturnafterunwind needs a VMException, got %s (%s)", REPR(ex)->name, MVM_6model_get_debug_name(tc, ex));
+}
